@@ -55,7 +55,7 @@ use RT::Shredder;
 
 package RT::Extension::MergeUsers;
 
-our $VERSION = '1.12';
+our $VERSION = '1.13';
 
 =head1 NAME
 
@@ -63,7 +63,7 @@ RT::Extension::MergeUsers - Merges two users into the same effective user
 
 =head1 RT VERSION
 
-Works with RT 4.0, 4.2, 4.4, 5.0, 6.0.
+Works with RT 5.0 (5.0.8 and newer), RT 6.0. For RT 4.0, 4.2, 4.4, 5.0 (up to 5.0.7) download version 1.11.
 
 =head1 DESCRIPTION
 
@@ -78,6 +78,111 @@ It also provides a version of L<CanonicalizeEmailAddress>, which means that
 all e-mail sent from secondary users is displayed as coming from the primary
 user.
 
+=head1 INSTALLATION
+
+Be sure to also read L</UPGRADING> if you are upgrading.
+
+=over
+
+=item C<perl Makefile.PL>
+
+=item C<make>
+
+=item C<make install>
+
+May need root permissions
+
+=item Edit your F</opt/rt6/etc/RT_SiteConfig.pm>
+
+Add this line:
+
+    Plugin('RT::Extension::MergeUsers');
+
+=item Clear your mason cache
+
+    rm -rf /opt/rt6/var/mason_data/obj
+
+=item Restart your webserver
+
+=back
+
+=head1 UPGRADING
+
+If you are upgrading from 0.03_01 or earlier, you must run
+F<bin/rt-update-merged-users>.  This script will create MergedUsers
+Attributes so RT can know when you're looking at a user that other users
+have been merged into. If you don't run this script, you'll have issues
+unmerging users. It can be safely run multiple times, it will only
+create Attributes as needed.
+
+=head1 UTILITIES
+
+=head2 rt-clean-merged-users
+
+When a user with another user merged into it is shredded,
+the attributes on that user are also shredded, but the
+merged user will remain, along with attributes that may point
+to the now missing user id. This script cleans up attributes
+if the merged-into user record is now gone. These users will then be
+converted back to regular unmerged users.
+
+=head2 rt-merge-users
+
+A command-line tool to merge one user into another
+
+=head1 CAVEATS
+
+=head2 RT::Shredder and Merged Users
+
+Merging a user effectively makes it impossible to load the merged user
+directly. Attempting to access the old user resolves to the merged-into user.
+Because of this, MergeUsers has some extra code to help L<RT::Shredder>
+clean up these merged records to avoid leaving merged user records in the DB
+while removing the user they were merged into.
+
+When running L<RT::Shredder> on a user record with other users merged into it,
+the merged users are Unmerged before the initial user record is shredded.
+There are two options to handle these newly unmerged users:
+
+=over
+
+=item 1.
+
+Re-run your shredder command with the same or similar options. The unmerged
+user records will now be accessible and, depending on your shredder options,
+they will likely be shredded on the second run. If you have multiple
+layers of merged users, you may need to run shredder multiple times.
+
+=item 2.
+
+MergeUsers will log the unmerged users at the C<info> level so you can pull
+the user ids from the log and shred them manually. This is most likely to
+be useful if you are shredding one specific user (and all merged accounts).
+
+=back
+
+=head2 rt-serializer
+
+MergeUsers is not compatible with C<rt-seralizer>, you need to disable the
+extension before running C<rt-serializer>.
+
+=head2 Single level of merged users
+
+A user can only be merged into a single user.
+
+A user can have multiple users merged into themselves.
+
+A user that has been merged into one user cannot be merged into a different
+user. You must first unmerge the user and then merge them into the different
+user.
+
+A user that has one or more users merged into themselves cannot be merged into
+another user. You must first unmerge all the merged users and then merge them
+all into the other user. Previous versions would allow multiple levels of
+merging when calling MergeInto from a command line script but some searching
+functionality does not work correctly in such cases and so it has been
+disallowed.
+
 =head1 REST2 API
 
 RT::Extension::MergeUsers provides REST2 API endpoints for programmatically
@@ -227,99 +332,6 @@ Example error response:
         "message": "User is a required field"
     }
 
-=head1 INSTALLATION
-
-Be sure to also read L</UPGRADING> if you are upgrading.
-
-=over
-
-=item C<perl Makefile.PL>
-
-=item C<make>
-
-=item C<make install>
-
-May need root permissions
-
-=item Edit your F</opt/rt6/etc/RT_SiteConfig.pm>
-
-If you are using RT 4.2 or greater, add this line:
-
-    Plugin('RT::Extension::MergeUsers');
-
-For RT 4.0, add this line:
-
-    Set(@Plugins, qw(RT::Extension::MergeUsers));
-
-or add C<RT::Extension::MergeUsers> to your existing C<@Plugins> line.
-
-=item Clear your mason cache
-
-    rm -rf /opt/rt6/var/mason_data/obj
-
-=item Restart your webserver
-
-=back
-
-=head1 UPGRADING
-
-If you are upgrading from 0.03_01 or earlier, you must run
-F<bin/rt-update-merged-users>.  This script will create MergedUsers
-Attributes so RT can know when you're looking at a user that other users
-have been merged into. If you don't run this script, you'll have issues
-unmerging users. It can be safely run multiple times, it will only
-create Attributes as needed.
-
-=head1 UTILITIES
-
-=head2 rt-clean-merged-users
-
-When a user with another user merged into it is shredded,
-the attributes on that user are also shredded, but the
-merged user will remain, along with attributes that may point
-to the now missing user id. This script cleans up attributes
-if the merged-into user record is now gone. These users will then be
-converted back to regular unmerged users.
-
-=head2 rt-merge-users
-
-A command-line tool to merge one user into another
-
-=head1 CAVEATS
-
-=head2 RT::Shredder and Merged Users
-
-Merging a user effectively makes it impossible to load the merged user
-directly. Attempting to access the old user resolves to the merged-into user.
-Because of this, MergeUsers has some extra code to help L<RT::Shredder>
-clean up these merged records to avoid leaving merged user records in the DB
-while removing the user they were merged into.
-
-When running L<RT::Shredder> on a user record with other users merged into it,
-the merged users are Unmerged before the initial user record is shredded.
-There are two options to handle these newly unmerged users:
-
-=over
-
-=item 1.
-
-Re-run your shredder command with the same or similar options. The unmerged
-user records will now be accessible and, depending on your shredder options,
-they will likely be shredded on the second run. If you have multiple
-layers of merged users, you may need to run shredder multiple times.
-
-=item 2.
-
-MergeUsers will log the unmerged users at the C<info> level so you can pull
-the user ids from the log and shred them manually. This is most likely to
-be useful if you are shredding one specific user (and all merged accounts).
-
-=back
-
-=head2 rt-serializer
-
-MergeUsers is not compatible with C<rt-seralizer>, you need to disable the
-extension before running C<rt-serializer>.
 
 =cut
 
@@ -327,19 +339,19 @@ package RT::User;
 
 our %EFFECTIVE_ID_CACHE;
 
-use RT::Interface::Web::Handler;
+sub EffectiveIDCacheNeedsUpdate {
+    my $self   = shift;
+    my $update = shift;
+    my $system = RT->System;
 
-{
-    my $i = 0;
-
-    my $old_cleanup = \&RT::Interface::Web::Handler::CleanupRequest;
-    no warnings 'redefine';
-    *RT::Interface::Web::Handler::CleanupRequest = sub {
-        $old_cleanup->(@_);
-        return if ++$i % 100; # flush cache every N requests
-        %EFFECTIVE_ID_CACHE = ();
-    };
+    if ($update) {
+        return $system->SetAttribute(Name => 'EffectiveIDCacheNeedsUpdate', Content => time);
+    } else {
+        my $cache = $system->FirstAttribute('EffectiveIDCacheNeedsUpdate');
+        return (defined $cache ? $cache->Content : 0 );
+    }
 }
+my $CACHE_TIME = RT::User->EffectiveIDCacheNeedsUpdate;
 
 sub CanonicalizeEmailAddress {
     my $self = shift;
@@ -367,6 +379,12 @@ sub LoadByCols {
     $self->SUPER::LoadByCols(@_);
     return $self->id unless my $oid = $self->id;
 
+    my $cache_time = RT::User->EffectiveIDCacheNeedsUpdate;
+    if ( $CACHE_TIME < $cache_time ) {
+        %EFFECTIVE_ID_CACHE = ();
+        $CACHE_TIME         = $cache_time;
+    }
+
     unless ( exists $EFFECTIVE_ID_CACHE{ $oid } ) {
         my $effective_id = RT::Attribute->new( $RT::SystemUser );
         $effective_id->LoadByCols(
@@ -446,14 +464,34 @@ sub MergeInto {
     return (0, "Could not merge @{[$merge->Name]} into itself")
            if $merge->id == $canonical_self->id;
 
+    # No merging if the user being merged has already been merged
+    my ($self_effective) = $canonical_self->Attributes->Named("EffectiveId");
+    return (0, "User @{[$canonical_self->Name]} has already been merged into @{[$self_effective->Content]}")
+           if defined $self_effective and $self_effective->Content;
+
     # No merging if the user you're merging into was merged into you
     # (ie. you're the primary address for this user)
     my ($new) = $merge->Attributes->Named("EffectiveId");
     return (0, "User @{[$canonical_self->Name]} has already been merged")
            if defined $new and $new->Content == $canonical_self->id;
 
+    # do not allow merging a user that has its own merged user(s)
+    my $self_merged_users = $canonical_self->FirstAttribute('MergedUsers');
+    if ( $self_merged_users && @{ $self_merged_users->Content } ) {
+        return (0, "User @{[$canonical_self->Name]} has merged users");
+    }
+
+    # If Privileged values for both users do not match, abort
+    my $merge_priv = $merge->Privileged // 0;
+    my $self_priv  = $self->Privileged // 0;
+    return ( 0,
+        "Cannot merge privileged users with unprivileged users, update the user's privileges first"
+        )
+        if $merge_priv != $self_priv;
+
     # clean the cache
     delete $EFFECTIVE_ID_CACHE{$self->id};
+    RT::User->EffectiveIDCacheNeedsUpdate(1);
 
     # do the merge
     $canonical_self->SetAttribute(
@@ -474,7 +512,7 @@ sub MergeInto {
         $canonical_self->Comments||'',
         "Merged into ". ($merge->EmailAddress || $merge->Name)." (". $merge->id .")",
     );
-    return (1, "Merged users successfuly");
+    return ($merge->id, "Merged users successfuly");
 }
 
 sub UnMerge {
@@ -486,6 +524,7 @@ sub UnMerge {
     # flush the cache, or the Sets below will
     # clobber $self
     delete $EFFECTIVE_ID_CACHE{$self->id};
+    RT::User->EffectiveIDCacheNeedsUpdate(1);
 
     my $merge = RT::User->new( $self->CurrentUser );
     $merge->Load( $current->Content );
@@ -969,7 +1008,7 @@ or via the web at
 
 =head1 LICENSE AND COPYRIGHT
 
-This software is Copyright (c) 2014-2025 by Best Practical Solutions
+This software is Copyright (c) 2014-2026 by Best Practical Solutions
 
 This is free software, licensed under:
 
